use serde::{Deserialize, Serialize};
use serde_json::json;
use sqlx::prelude::FromRow;
use std::collections::HashSet;

use crate::{
    get_pg,
    library::{album::Album, artist::Artist},
};

use super::{event::Event, metadata::AudioMetadata, user::User};

#[derive(Debug, Clone, Serialize, Deserialize, FromRow)]
pub struct Track {
    pub id: uuid::Uuid,
    pub path: String,
    pub title: String,
    pub date_added: chrono::DateTime<chrono::Utc>,
    pub album: Option<uuid::Uuid>,
    pub artist: Option<uuid::Uuid>,
    pub meta: Option<serde_json::Value>,
}

impl Track {
    pub async fn create(data: &serde_json::Map<String, serde_json::Value>) {
        sqlx::query("INSERT INTO track (path, title, meta) VALUES ($1, $2, $4)")
            .bind(data.get("path").unwrap().as_str().unwrap().to_string())
            .bind(data.get("title").unwrap().as_str().unwrap().to_string())
            .bind(data.get("meta"))
            .fetch(get_pg!());
    }

    pub async fn get_latest_of_user(u: &User) -> Vec<Self> {
        let latest_events = Event::get_latest_events_of(u).await;
        let mut ids = HashSet::new();

        for event in latest_events {
            if !ids.contains(&event.track) {
                ids.insert(event.track.clone());
            }
        }

        sqlx::query_as("SELECT * FROM track WHERE id = ANY($1)")
            .bind(ids.into_iter().collect::<Vec<_>>())
            .fetch_all(get_pg!())
            .await
            .unwrap()
    }

    /// Transcode audio to OPUS with `bitrate`
    pub fn get_opus(&self, bitrate: u32) -> Option<String> {
        self.transcode("libopus", bitrate, "opus")
    }

    /// Transcode audio to AAC with `bitrate`
    pub fn get_aac(&self, bitrate: u32) -> Option<String> {
        self.transcode("aac", bitrate, "aac")
    }

    /// Transcode audio
    pub fn transcode(&self, codec: &str, bitrate: u32, ext: &str) -> Option<String> {
        let transcoded = format!("./data/transcode/{codec}/{bitrate}/{}.{ext}", self._id);

        if std::path::Path::new(&transcoded).exists() {
            return Some(transcoded);
        }

        log::info!("Transcoding {} to {} {}", self._id, codec, bitrate);

        std::fs::create_dir_all(format!("./data/transcode/{codec}/{bitrate}")).unwrap();

        let out = std::process::Command::new("ffmpeg")
            .arg("-i")
            .arg(&self.path)
            .arg("-c:a")
            .arg(codec)
            .arg("-b:a")
            .arg(format!("{bitrate}k"))
            .arg(&transcoded)
            .output()
            .unwrap();

        if !out.status.success() {
            return None;
        }

        Some(transcoded)
    }

    /// Find tracks with no album or artist
    pub async fn get_orphans() -> Vec<Track> {
        // todo : fix
        Self::find(
            doc! {
                "artist_id": None::<String>,
                "album_id": None::<String>
            },
            None,
            None,
        )
        .await
        .unwrap()
    }
}

impl Track {
    async fn api(&self) -> serde_json::Value {
        // todo : fix
        let (cover, album_title, album_id) = if let Some(album_ref) = &self.album {
            let album = album_ref.get::<Album>().await;

            (album.get_cover().await.is_some(), album.title, album._id)
        } else {
            (false, String::new(), String::new())
        };

        let artist_title = if let Some(artist_ref) = &self.artist {
            artist_ref
                .get_partial::<Artist>(json!({"name": 1}))
                .await
                .name
        } else {
            None
        };

        json!({
            "id": self.id,
            "title": self.title,
            "track_number": self.meta.as_ref().map(|x| AudioMetadata(*x).track_number()),
            "meta": serde_json::to_value(&self.meta).unwrap(),
            "album_id": self.album,
            "album": album_title,
            "cover": if cover {
                Some(format!("/album/{album_id}/cover"))
            } else {
                None
            },
            "artist_id": self.artist,
            "artist": artist_title
        })
    }
}
